import numpy as np
from matplotlib import pyplot as plt
import os


def plot_distribution(w, bins=50, title=None, figsize=(8, 6), xlabel=r"$w$", **kwargs):
    """Plot the distribution of a vector of weights."""
    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot(111)
    ax.hist(w, bins=bins, **kwargs)
    ax.set_xlabel(xlabel)
    # ax.set_ylabel('frequency')
    if title is not None:
        ax.set_title(title)
    return fig


def plot_lorenz_curve(
    w_set,
    labels,
    dpi=150,
    savefile=None,
    plot_equality=True,
    plot_uniform=True,
    **kwargs,
):
    """
    Plots the Lorenz curve for a set of wealth distributions.

    Parameters:
    - w_set (list): List of wealth distributions to plot.
    - labels (list): List of labels for each wealth distribution.
    - dpi (int, optional): Dots per inch of the plot. Defaults to 150.
    - savefile (str, optional): Filepath to save the plot. Defaults to None.
    - plot_equality (bool, optional): Whether to plot the equality curve. Defaults to True.
    - plot_uniform (bool, optional): Whether to plot the uniform distribution curve. Defaults to True.
    - **kwargs: Additional keyword arguments to customize the plot.

    Returns:
    None
    """
    fig, ax = plt.subplots(dpi=dpi)

    ax.set_ylabel("Cumulative wealth")
    ax.set_xlabel("Fraction of agents")
    ax.set_xticks(np.arange(0, 1.05, 0.05), minor=True)
    ax.set_yticks(np.arange(0, 1.05, 0.05), minor=True)

    if plot_equality:
        # Plot gini=0 curve
        ax.plot([0, 1], [0, 1], color="k", label="Total equality")
    if plot_uniform:
        # Plot uniform distribution curve
        agent = np.random.rand(10000)
        # Normalize wealth
        agent /= np.sum(agent)
        agent = np.sort(agent)
        ax.plot(
            np.linspace(0, 1, agent.shape[0]),
            np.cumsum(agent) / np.sum(agent),
            label="Initial condition",
            color="black",
            linestyle="--",
        )

    for w, label in zip(w_set, labels):

        w = np.sort(w)

        ax.plot(
            np.linspace(0, 1, w.shape[0]),
            np.cumsum(w) / np.sum(w),
            label=label,
        )
        ax.grid(which="minor", alpha=0.5, linestyle="--")
        ax.grid(which="major", alpha=1, linestyle="-")

    ax.legend(fontsize=8)

    plt.show()
    if savefile:
        fig.savefig(os.path.join(savefile, "lorenz_curve"), format="png")


def plot_mean_wealth_by_risk(
    wealths,
    risks,
    r_min=0,
    r_max=1,
    num_bins=100,
    error_bars=None,
    log=None,
    title=None,
):
    """
    Plots the mean wealth as a function of risk.

    Args:
        wealths (list): List of wealth values.
        risks (list): List of risk values.
        r_min (float, optional): Minimum risk value. Defaults to 0.
        r_max (float, optional): Maximum risk value. Defaults to 1.
        num_bins (int, optional): Number of bins for risk values. Defaults to 100.
        title (str, optional): Title of the plot. Defaults to None.

    Returns:
        tuple: A tuple containing the mean risk values and the mean wealth values for each bin.
    """
    # Compute the bin edges
    bin_edges = np.linspace(r_min, r_max, num_bins + 1)

    # Group wealth values (wealths) by binned risks and calculate mean wealth for each bin
    mean_wealth_by_bin = []
    std_wealth_by_bin = []
    for i in range(num_bins):
        # Filter wealth values corresponding to the current risk bin
        bin_mask = (risks >= bin_edges[i]) & (risks < bin_edges[i + 1])

        wealths_in_bin = [w for i, w in enumerate(wealths) if bin_mask[i]]

        # Calculate mean wealth for the current risk bin
        mean_wealth = np.mean(wealths_in_bin if len(wealths_in_bin) > 10 else 0)
        std_wealth = np.std(wealths_in_bin if len(wealths_in_bin) > 10 else 0)

        mean_wealth_by_bin.append(mean_wealth)
        std_wealth_by_bin.append(std_wealth / np.sqrt(1000))

    # Calculate the mean risk value for each bin
    mean_risk_by_bin = [(bin_edges[i] + bin_edges[i + 1]) / 2 for i in range(num_bins)]

    # Find the index of the bin with maximum wealth
    max_wealth_index = np.argmax(mean_wealth_by_bin)

    # Retrieve the mean risk and mean wealth for the bin with maximum wealth
    max_wealth_bin_risk = mean_risk_by_bin[max_wealth_index]
    max_wealth_bin_wealth = mean_wealth_by_bin[max_wealth_index]

    # Print the mean risk and mean wealth for the bin with maximum wealth
    print(
        f"El bin con riesgo medio: {max_wealth_bin_risk:.3f} es el que tiene el maximo de riqueza media: {max_wealth_bin_wealth:.5f}"
    )

    # Plot mean wealth +- std for each risk bin and mark the bin with maximum wealth
    plt.figure(figsize=(10, 6))
    plt.plot(mean_risk_by_bin, mean_wealth_by_bin)
    if error_bars:
        plt.fill_between(
            mean_risk_by_bin,
            np.array(mean_wealth_by_bin)
            - np.array(std_wealth_by_bin) / np.sqrt(error_bars),
            np.array(mean_wealth_by_bin)
            + np.array(std_wealth_by_bin) / np.sqrt(error_bars),
            alpha=0.3,
        )
    plt.scatter(
        max_wealth_bin_risk, max_wealth_bin_wealth, color="red", label="Max wealth bin"
    )
    plt.xlim(0, r_max + 0.1)
    plt.xlabel("Riesgo medio del bin")
    plt.ylabel("Riqueza promedio")

    if log:
        plt.yscale("log")
    if not title:
        plt.title(f"Riqueza promedio del bin vs Riesgo ({num_bins} bines)")
    else:
        plt.title(title)
    plt.grid(True)
    plt.show()

    return mean_risk_by_bin, mean_wealth_by_bin



import math
from itertools import combinations

import networkx as nx
from networkx.utils import py_random_state


# Generating function for additive and multiplicative threshold graphs,
# slightly modified version from the function provided in the networkx package.
@py_random_state(7)
def geographical_threshold_graph_custom(
    n,
    theta,
    dim=2,
    pos=None,
    weight=None,
    metric=None,
    p_dist=None,
    seed=None,
    join="add",
):
    r"""Returns a geographical threshold graph.

    The geographical threshold graph model places $n$ nodes uniformly at
    random in a rectangular domain.  Each node $u$ is assigned a weight
    $w_u$. Two nodes $u$ and $v$ are joined by an edge if

    .. math::

       - Additive type:
       (w_u + w_v)p_{dist}(r) \ge \theta

       - Multiplicative type
       (w_u * w_v)p_{dist}(r) \ge \theta

    where `r` is the distance between `u` and `v`, `p_dist` is any function of
    `r`, and :math:`\theta` as the threshold parameter. `p_dist` is used to
    give weight to the distance between nodes when deciding whether or not
    they should be connected. The larger `p_dist` is, the more prone nodes
    separated by `r` are to be connected, and vice versa.

    Parameters
    ----------
    n : int or iterable
        Number of nodes or iterable of nodes
    theta: float
        Threshold value
    dim : int, optional
        Dimension of graph
    pos : dict
        Node positions as a dictionary of tuples keyed by node.
    weight : dict
        Node weights as a dictionary of numbers keyed by node.
    metric : function
        A metric on vectors of numbers (represented as lists or
        tuples). This must be a function that accepts two lists (or
        tuples) as input and yields a number as output. The function
        must also satisfy the four requirements of a `metric`_.
        Specifically, if $d$ is the function and $x$, $y$,
        and $z$ are vectors in the graph, then $d$ must satisfy

        1. $d(x, y) \ge 0$,
        2. $d(x, y) = 0$ if and only if $x = y$,
        3. $d(x, y) = d(y, x)$,
        4. $d(x, z) \le d(x, y) + d(y, z)$.

        If this argument is not specified, the Euclidean distance metric is
        used.

        .. _metric: https://en.wikipedia.org/wiki/Metric_%28mathematics%29
    p_dist : function, optional
        Any function used to give weight to the distance between nodes when
        deciding whether or not they should be connected. `p_dist` was
        originally conceived as a probability density function giving the
        probability of connecting two nodes that are of metric distance `r`
        apart. The implementation here allows for more arbitrary definitions
        of `p_dist` that do not need to correspond to valid probability
        density functions. The :mod:`scipy.stats` package has many
        probability density functions implemented and tools for custom
        probability density definitions, and passing the ``.pdf`` method of
        scipy.stats distributions can be used here. If ``p_dist=None``
        (the default), the exponential function :math:`r^{-2}` is used.
    seed : integer, random_state, or None (default)
        Indicator of random number generation state.
        See :ref:`Randomness<randomness>`.

    Returns
    -------
    Graph
        A random geographic threshold graph, undirected and without
        self-loops.

        Each node has a node attribute ``pos`` that stores the
        position of that node in Euclidean space as provided by the
        ``pos`` keyword argument or, if ``pos`` was not provided, as
        generated by this function. Similarly, each node has a node
        attribute ``weight`` that stores the weight of that node as
        provided or as generated.

    Examples
    --------
    Specify an alternate distance metric using the ``metric`` keyword
    argument. For example, to use the `taxicab metric`_ instead of the
    default `Euclidean metric`_::

        >>> dist = lambda x, y: sum(abs(a - b) for a, b in zip(x, y))
        >>> G = nx.geographical_threshold_graph(10, 0.1, metric=dist)

    .. _taxicab metric: https://en.wikipedia.org/wiki/Taxicab_geometry
    .. _Euclidean metric: https://en.wikipedia.org/wiki/Euclidean_distance

    Notes
    -----
    If weights are not specified they are assigned to nodes by drawing randomly
    from the exponential distribution with rate parameter $\lambda=1$.
    To specify weights from a different distribution, use the `weight` keyword
    argument::

    >>> import random
    >>> n = 20
    >>> w = {i: random.expovariate(5.0) for i in range(n)}
    >>> G = nx.geographical_threshold_graph(20, 50, weight=w)

    If node positions are not specified they are randomly assigned from the
    uniform distribution.

    References
    ----------
    .. [1] Masuda, N., Miwa, H., Konno, N.:
       Geographical threshold graphs with small-world and scale-free
       properties.
       Physical Review E 71, 036108 (2005)
    .. [2]  Milan Bradonjić, Aric Hagberg and Allon G. Percus,
       Giant component and connectivity in geographical threshold graphs,
       in Algorithms and Models for the Web-Graph (WAW 2007),
       Antony Bonato and Fan Chung (Eds), pp. 209--216, 2007
    """
    G = nx.empty_graph(n)
    # If no weights are provided, choose them from an exponential
    # distribution.
    if weight is None:
        weight = {v: seed.expovariate(1) for v in G}
    # If no positions are provided, choose uniformly random vectors in
    # Euclidean space of the specified dimension.
    if pos is None:
        pos = {v: [seed.random() for i in range(dim)] for v in G}
    # If no distance metric is provided, use Euclidean distance.
    if metric is None:
        metric = math.dist
    nx.set_node_attributes(G, weight, "weight")
    nx.set_node_attributes(G, pos, "pos")

    # if p_dist is not supplied, use default r^-2
    if p_dist is None:
        def p_dist(r):
            return r**-2

    # Returns ``True`` if and only if the nodes whose attributes are
    # ``du`` and ``dv`` should be joined, according to the threshold
    # condition.
    def should_join(pair):
        u, v = pair
        u_pos, v_pos = pos[u], pos[v]
        u_weight, v_weight = weight[u], weight[v]

        # Lautaro Modified. If not additive, weights join multiplicatively
        if join == "add":
            return (u_weight + v_weight) * p_dist(metric(u_pos, v_pos)) >= theta

        if join == "mul":
            return (u_weight * v_weight) * p_dist(metric(u_pos, v_pos)) >= theta

        else:
            raise TypeError("join should be 'add' or 'mul'.")

    G.add_edges_from(filter(should_join, combinations(G, 2)))
    return G



# Agregados por gpt
from concurrent.futures import ThreadPoolExecutor
from tqdm import tqdm

class GPUModel:
    def __init__(self, n_agents=100, G=None, f=0, w_min=3e-17, w_0=None, r_min=0, r_max=1, tpb=32, bpg=512, stream=None, seed=None):
        self.n_agents = n_agents
        self.w_min = w_min
        self.G = G

        np.random.seed(seed)
        self.w = np.random.rand(self.n_agents).astype(np.float32)
        self.r = np.random.uniform(r_min, r_max, self.n_agents).astype(np.float32)
        self.m = np.zeros((self.n_agents), dtype=np.int32)
        self.f = f
        self.w = w_0 if w_0 is not None else self.w / np.sum(self.w)
        
        if self.G is not None:
            self.c_neighs, self.neighs = self.G.get_neighbors_array_gpu()

        self.stream = stream if stream is not None else cuda.default_stream()
        self.tpb = tpb
        self.bpg = bpg
        random_seed = seed if seed is not None else np.random.randint(0, 0x7FFFFFFFFFFFFFFF)
        self.rng_state = create_xoroshiro128p_states(self.n_agents, seed=random_seed)

    def MCS(self, steps: int, tpb: int = None, bpg: int = None, rng_state: np.ndarray = None) -> None:
        tpb = tpb or self.tpb
        bpg = bpg or self.bpg
        rng_state = rng_state or self.rng_state
        
        with cuda.pinned(self.w), cuda.pinned(self.r), cuda.pinned(self.m):
            w_d = cuda.to_device(self.w, stream=self.stream)
            r_d = cuda.to_device(self.r, stream=self.stream)
            m_d = cuda.to_device(self.m, stream=self.stream)

            if self.G is None:
                k_ys.k_ys_mcs[bpg, tpb, self.stream](self.n_agents, w_d, r_d, m_d, self.w_min, self.f, steps, rng_state)
            else:
                c_neighs_d = cuda.to_device(self.c_neighs, stream=self.stream)
                neighs_d = cuda.to_device(self.neighs, stream=self.stream)
                k_ys.k_ys_mcs_graph[bpg, tpb, self.stream](self.n_agents, w_d, r_d, m_d, c_neighs_d, neighs_d, self.w_min, self.f, steps, rng_state)
                del c_neighs_d, neighs_d

            w_d.copy_to_host(self.w, self.stream)
        del w_d, r_d, m_d
        cuda.synchronize()


class GPUEnsemble:
    def __init__(self, n_models=1, n_agents=1000, tpb=32, bpg=512, graphs=None, **kwargs):
        self.n_streams = n_models
        self.n_agents = n_agents
        self.tpb = tpb
        self.bpg = bpg
        self.graphs = graphs
        random_seeds = np.random.randint(0, 0x7FFFFFFFFFFFFFFF, size=self.n_streams)

        if self.n_streams == 1:
            self.streams = [cuda.default_stream()]
        else:
            self.streams = [cuda.stream() for _ in range(self.n_streams)]

        if self.graphs is None:
            self.models = [GPUModel(n_agents=n_agents, stream=stream, **kwargs) for stream in self.streams]
        else:
            self.models = [GPUModel(n_agents=n_agents, stream=stream, G=self.graphs[i], **kwargs) for i, stream in enumerate(self.streams)]

        self.rng_states = [create_xoroshiro128p_states(n_agents, seed=random_seeds[i]) for i in range(self.n_streams)]

    def MCS(self, steps: int, verbose: bool = False) -> None:
        with ThreadPoolExecutor() as executor:
            futures = [executor.submit(model.MCS, steps, self.tpb, self.bpg, rng_state) for model, rng_state in zip(self.models, self.rng_states)]
            for future in tqdm(futures):
                future.result()

    def save_wealths(self, filepath: str = None) -> None:
        if filepath is None:
            raise ValueError("Insert a valid filepath with the structure 'path/name'")
        wealths = np.array([[self.models[i].w] for i in range(self.n_streams)])
        np.save(filepath, wealths)

    def get_gini(self) -> tuple:
        ginis = [measures.gini(model.w) for model in self.models]
        return np.mean(ginis), np.std(ginis)

    def get_n_active(self) -> tuple:
        n_active = [measures.num_actives(model.w, model.w_min) for model in self.models]
        return np.mean(n_active), np.std(n_active)

    def get_n_frozen(self) -> tuple:
        n_frozen = [measures.num_frozen(model.w, model.w_min, model.G) for model in self.models]
        return np.mean(n_frozen), np.std(n_frozen)


from .graphs.custom_gtg import *
from .graphs.graph_class import *
from .kernels.k_ys import *
from .kernels.locks import *
from .metrics.measures import *
from .metrics.gpu_measures import *
from .models.model import *
from .trades.ys import *
from .utils.plots import *


def yard_sale(ri, wi, rj, wj):
    """
    Calculate the Yard-Sale interaction between two agents.

    Parameters:
    ri (int): Risk value of the first agent.
    wi (int): Wealth value of the first agent.
    rj (int): Risk value of the second agent.
    wj (int): Wealth value of the second agent.

    Returns:
    int: The minimum value between ri * wi and rj * wj.
    """
    return min(ri * wi, rj * wj)
